__author__ = 'maartenbreddels'

import time

import vaex.ui.plugin
from vaex.ui.qt import *


#import vaex.plot_windows

import logging
#import vaex.ui.undo as undo

logger = logging.getLogger("vaex.ui.plugin.animation")

class AnimationPlugin(vaex.ui.plugin.PluginLayer):
	name = "animation"
	def __init__(self, parent, layer):
		super(AnimationPlugin, self).__init__(parent, layer)
		layer.plug_page(self.plug_page, "Animation", 6.0, 1.0)

		self.timer_sequence = QtCore.QTimer(parent)
		self.timer_sequence.timeout.connect(self.on_timeout_sequence)
		self.timer_sequence.setInterval(40) # max 50 fps to save cpu/battery?
		self.has_snapshots = self.dataset.has_snapshots()

		self.no_snapshots = self.dataset.rank1s[list(self.dataset.rank1s.keys())[0]].shape[0] if self.has_snapshots else 0
		#self.plot_original = self.dialog.plot
		#self.dialog.plot = self.plot_wrapper


		def on_plot_finished(plot_window, figure):
			#if self.timer_sequence.isActive():
			#	self.dialog.update_grids()
			#else:
			if self.record_frames:
				index = self.dataset.selected_serie_index
				path = self.frame_template.format(index=index)
				figure.savefig(path)
				if index == self.no_snapshots-1:
					self.record_frames = False
					self.frame_template = None

		self.layer.plot_window.signal_plot_finished.connect(on_plot_finished)

		self.record_frames = False
		self.frame_template = None
		#self.dataset.serie_index_selection_listeners.append(self.on_snapshot_select)
		# instead of binding to the event listener, wrap it so we are sure we execute it afterwards
		# this assumes the function isn't bound yet
		#def wrapper(index, previous=self.dialog.onSerieIndexSelect):
		#	previous(index)
		#	print ">>>>3", index, self.dataset.selected_serie_index
		#	self.box_sequence.setCurrentIndex(self.dataset.selected_serie_index)
		#self.dialog.onSerieIndexSelect = wrapper
		def on_sequence_index_change(dataset, index):
			self.box_sequence.setCurrentIndex(index)
		self.dataset.signal_sequence_index_change.connect(on_sequence_index_change)
		#print "set time to zero" * 100
		self.dataset.variables["time"] = str(0.)


		self.timer_realtime = QtCore.QTimer(parent)
		self.timer_realtime.timeout.connect(self.on_timeout_realtime)
		self.timer_realtime.setInterval(40) # max 50 fps to save cpu/battery?
		self.time_start = time.time()

	def clean_up(self):
		self.timer_sequence.stop()
		self.timer_realtime.stop()
		#self.dataset.serie_index_selection_listeners.remove(self.on_snapshot_select)

	def on_timeout_realtime(self):
		self.dataset.variables["time"] = str(self.time_start - time.time())
		#self.dialog.compute()
		#self.layer.jobs_manager.execute()


	def on_timeout_sequence(self):
		#self.dialog.update_grids()
		#if self.
		index = self.dataset.selected_serie_index + 1
		if index < self.no_snapshots:
			msg = "snapshot %d/%d" % (index+1, self.no_snapshots)
			self.layer.plot_window.message(msg, index=-1)
			self.dataset.selectSerieIndex(index)
			for layer in self.layer.plot_window.layers:
				if layer.dataset != self.dataset:
					layer.dataset.selectSerieIndex(index)
			#$self.layer.
			self.layer.jobs_manager.execute()
		else:
			self.layer.plot_window.message(None, index=-1)
			self.timer_sequence.stop()
			self.stop_animation()


	def plug_page(self, page):
		layout = self.layout = QtGui.QGridLayout()
		page.setLayout(self.layout)
		layout.setSpacing(0)
		layout.setContentsMargins(0,0,0,0)
		layout.setAlignment(QtCore.Qt.AlignTop)

		row = 0

		has_snapshots = self.dataset.has_snapshots()



		self.group_box_sequence = QtGui.QGroupBox("Snapshot animation", page)
		layout.addWidget(self.group_box_sequence, row, 1)
		row += 1
		layout_sequence = self.layout_sequence = QtGui.QGridLayout()
		#self.group_box_sequence.setLayout(self.layout_sequence)
		layout_sequence.setSpacing(0)
		layout_sequence.setContentsMargins(0,0,0,0)
		layout_sequence.setAlignment(QtCore.Qt.AlignTop)
		row_sequence = 0


		self.layout_control = QtGui.QHBoxLayout()
		self.layout_control.setSpacing(0)
		logger.debug("3")

		def add_control_button(text, handler):
			button = QtGui.QToolButton(self.group_box_sequence)
			button.setText(text)
			#button.setAutoDefault(False)
			button.setContentsMargins(0, 0, 0, 0)
			button.setEnabled(has_snapshots)
			#button.setFlat(True)
			self.layout_control.addWidget(button, 1)
			def wrap_handler():
				handler()
				self.layer.jobs_manager.execute()

			button.clicked.connect(wrap_handler)
			
		def select(index):
			self.dataset.selectSerieIndex(index)
			for layer in self.layer.plot_window.layers:
				if layer.dataset != self.dataset:
					layer.dataset.selectSerieIndex(index)
			#$self.layer.
			self.layer.jobs_manager.execute()
		def do_begin():
			select(0)
		def do_end():
			select(self.no_snapshots-1)
		def do_prev():
			select(max(0, self.dataset.selected_serie_index-1))
		def do_next():
			select(min(self.no_snapshots-1, self.dataset.selected_serie_index+1))
		def do_prev_ten():
			select(max(0, self.dataset.selected_serie_index-10))
		def do_next_ten():
			select(min(self.no_snapshots-1, self.dataset.selected_serie_index+10))
		add_control_button("|<", do_begin)
		add_control_button("<<", do_prev_ten)
		add_control_button("<", do_prev)
		add_control_button(">", do_next)
		add_control_button(">>", do_next_ten)
		add_control_button(">|", do_end)

		self.layout_sequence.addLayout(self.layout_control, row_sequence, 1)
		row_sequence += 1
		#self.label_sequence_info = QtGui.QLabel("""
#A snapshot animation is an animation where
#each frame shows a different snapshot,
#where usually the next snapshot is a next moment in time).
#		""".strip(), self.group_box_sequence)
		#self.layout_sequence.addWidget(self.label_sequence_info, row_sequence, 1)
		row_sequence += 1

		self.box_sequence = QtGui.QComboBox(self.group_box_sequence)
		self.box_sequence.addItems([str(k) for k in range(self.no_snapshots)])
		self.box_sequence.setCurrentIndex(self.dataset.current_sequence_index())
		self.box_sequence.setEnabled(has_snapshots)
		def on_sequence_change(index):
			if index != self.dataset.current_sequence_index():
				self.dataset.selectSerieIndex(index)
				for layer in self.layer.plot_window.layers:
					if layer.dataset != self.dataset:
						layer.dataset.selectSerieIndex(index)
				#self.dataset.selectSerieIndex(index)
				self.layer.jobs_manager.execute()
		self.box_sequence.currentIndexChanged.connect(on_sequence_change)
		self.layout_sequence.addWidget(self.box_sequence, row_sequence, 1)
		row_sequence += 1

		self.button_play_sequence = QtGui.QPushButton("Play(snapshots)", self.group_box_sequence)
		self.button_play_sequence.setCheckable(True)
		self.button_play_sequence.setAutoDefault(False)
		self.button_play_sequence.setEnabled(has_snapshots)
		if has_snapshots:
			self.button_play_sequence.setToolTip("This data set has no snapshots")
		layout_sequence.addWidget(self.button_play_sequence, row_sequence, 1)
		row_sequence += 1

		def on_toggle_play_sequence(checked):
			if checked:
				self.start_animation(self.button_play_sequence)
				#self.dataset.selectSerieIndex(0)
				self.layer.jobs_manager.execute()
				self.timer_sequence.start()
			else:
				self.timer_sequence.stop()
				self.stop_animation()
		self.button_play_sequence.toggled.connect(on_toggle_play_sequence)


		self.button_record_sequence = QtGui.QPushButton("Record(snapshots)", self.group_box_sequence)
		self.button_record_sequence.setCheckable(True)
		self.button_record_sequence.setAutoDefault(False)
		has_snapshots = self.dataset.has_snapshots()
		self.button_record_sequence.setEnabled(has_snapshots)
		if has_snapshots:
			self.button_record_sequence.setToolTip("This data set has no snapshots")
		layout_sequence.addWidget(self.button_record_sequence, row_sequence, 1)
		row_sequence += 1

		def on_toggle_record_sequence(checked):
			self.record_frames = False
			if checked:
				directory = getdir(self.parent, "Choose where to save frames", "")
				if directory:
					self.frame_template = os.path.join(directory, "%s_{index:05}.png" % self.dataset.name)
					self.frame_template = gettext(self.parent, "template for frame filenames", "template:", self.frame_template)
					if self.frame_template:
						self.record_frames = True
						#self.dataset.selectSerieIndex(0)
						self.layer.jobs_manager.execute()
						self.start_animation(self.button_record_sequence)
						self.timer_sequence.start()
			else:
				self.timer_sequence.stop()
				self.stop_animation()
		self.button_record_sequence.toggled.connect(on_toggle_record_sequence)


		self.group_box_realtime = QtGui.QGroupBox("Realtime animation", page)
		layout.addWidget(self.group_box_realtime, row, 1)
		layout_realtime = self.layout_realtime = QtGui.QGridLayout()
		self.group_box_realtime.setLayout(self.layout_realtime)
		layout_realtime.setSpacing(0)
		layout_realtime.setContentsMargins(0,0,0,0)
		layout_realtime.setAlignment(QtCore.Qt.AlignTop)
		row_sequence = 0
		self.button_play_realtime = QtGui.QPushButton("Play(time)", self.group_box_sequence)
		self.button_play_realtime.setCheckable(True)
		self.button_play_realtime.setAutoDefault(False)
		if has_snapshots:
			self.button_play_realtime.setToolTip("This data set has no snapshots")
		layout_realtime.addWidget(self.button_play_realtime, row_sequence, 1)
		row_sequence += 1

		def on_toggle_play_realtime(checked):
			if checked:
				self.start_animation(self.button_play_realtime)
				self.time_start = time.time()
				self.timer_realtime.start()
			else:
				self.timer_realtime.stop()
				self.stop_animation()
			self.button_play_realtime.toggled.connect(on_toggle_play_realtime)



	def start_animation(self, source_widget):
		widgets = [self.button_play_sequence, self.button_record_sequence, self.box_sequence]
		for widget in widgets:
			if source_widget != widget:
				widget.setEnabled(False)

	def stop_animation(self):
		widgets = [self.button_play_sequence, self.button_record_sequence, self.box_sequence]
		for widget in widgets:
			widget.setEnabled(True)

		self.button_play_sequence.setChecked(False)
		self.button_record_sequence.setChecked(False)




import logging
import numpy as np
import copy

logger = logging.getLogger("vaex.ui.undo")


class UndoManager(object):
    def __init__(self, max_bytes=1024**3):
        self.actions_undo = []
        self.actions_redo = []
        self.undo_count = 0  # number of times undo is pressed

    def undo(self):
        logger.debug("history was %r-%r" % (self.actions_undo, self.actions_redo))
        action = self.actions_undo.pop()
        logger.debug("undoing: %r" % action)
        self.actions_redo.insert(0, action)
        try:
            action.undo()
        except:
            logger.exception("error executing action")
        logger.debug("history is  %r-%r" % (self.actions_undo, self.actions_redo))

    def add_action(self, action):
        # cut off any remaining 'redo' action, and add action to the list
        logger.debug("history was %r-%r" % (self.actions_undo, self.actions_redo))
        logger.debug("adding action: %r" % action)
        self.actions_redo = []
        self.actions_undo.append(action)
        logger.debug("history is  %r-%r" % (self.actions_undo, self.actions_redo))

    def redo(self):
        logger.debug("history was %r-%r" % (self.actions_undo, self.actions_redo))
        logger.debug("redoing")
        action = self.actions_redo.pop(0)
        try:
            action.do()
        except:
            logger.exception("error executing action")
        self.actions_undo.append(action)
        logger.debug("history is  %r-%r" % (self.actions_undo, self.actions_redo))

    def can_undo(self):
        return len(self.actions_undo) > 0

    def can_redo(self):
        return len(self.actions_redo) > 0


class Action(object):
    """
    action should support
     - byteSize() # nr of bytes the action occupies
     - do() - does the operation, used at moment of the actual action, and during redo
     - redo() - redo the operation
     - description() - gui friendly desciption

     and actions should add itself to it's UndoManager
    """
    pass


class ActionMask(Action):
    def __init__(self, undo_manager, description, mask, apply_mask):
        """ Assuming mask is a bool array"""
        # store the 1 byte mask as a 1 bit mask to save memory
        self.undo_manager = undo_manager
        self.data = None if mask is None else np.packbits(mask.astype(np.uint8))
        self.length = 0 if mask is None else len(mask)
        self._description = description
        self.mask = mask
        self.apply_mask = apply_mask
        self.undo_manager.add_action(self)

    def description(self):
        return self._description

    def do(self):
        mask = None if self.data is None else np.unpackbits(self.data).astype(np.bool)[:self.length]
        self.apply_mask(mask)

    def undo(self):
        # find a previous ActionMask, and execute it, but just keep it in the history
        for action in self.undo_manager.actions_undo[::-1]:  # traverse from most recent to last
            if isinstance(action, ActionMask):
                action.do()
                break
        else:  # if nothing was selected before, select None
            self.apply_mask(None)


class ActionZoom(Action):
    def __init__(self, undo_manager, description, apply_ranges, all_axis_indices,
                 previous_ranges_viewport, previous_range_level_show, axis_indices, ranges_viewport=None, range_level_show=None):
        self.undo_manager = undo_manager
        self.apply_ranges = apply_ranges

        self.all_axis_indices = all_axis_indices
        # self.previous_ranges = list(previous_ranges)
        self.previous_ranges_viewport = copy.deepcopy(previous_ranges_viewport)
        self.previous_range_level_show = None if previous_range_level_show is None else copy.deepcopy(previous_range_level_show)

        self.axis_indices = axis_indices
        # self.ranges = ranges
        self.ranges_viewport = copy.deepcopy(ranges_viewport)
        self.range_level_show = copy.deepcopy(range_level_show)
        self._description = description
        self.undo_manager.add_action(self)

    def description(self):
        return self._description

    def do(self):
        self.apply_ranges(self.axis_indices, self.ranges_viewport, self.range_level_show)

    def undo(self):
        self.apply_ranges(self.all_axis_indices, self.previous_ranges_viewport, self.previous_range_level_show)



import functools

import matplotlib.widgets

import vaex.ui.plugin
from vaex.ui import undo
from vaex.ui.qt import *
from vaex.ui.icons import iconfile
import logging
import vaex.ui.undo as undo
import vaex.ui.qt as dialogs
import re
logger = logging.getLogger("vaex.plugin.tasks")

def loglog(plot_window):
	for layer in plot_window.layers:
		for i, expression in enumerate(layer.state.expressions):
			layer.set_expression("log10(%s)" % expression, i)
	plot_window.queue_history_change("task: log/log")

def removelog(plot_window):
	def remove_log(expression):
		return re.sub("^\s*(log|log2|log10)\((.*?)\)\s*$", "\\2", expression)
	for layer in plot_window.layers:
		for i, expression in enumerate(layer.state.expressions):
			layer.set_expression(remove_log(expression), i)
	plot_window.queue_history_change("task: remove log/log")

def loglog_and_sigma3(plot_window):
	removelog(plot_window)
	loglog(plot_window)
	plot_window.queue_history_change(None)
	sigma3(plot_window)
	plot_window.queue_history_change("task: log/log and 3 sigma region")

def sigma3(plot_window):
	if plot_window.layers:
		layer = plot_window.layers[0]


		if layer.dataset.is_local():
			executor = vaex.execution.Executor()
		else:
			executor = vaex.remote.ServerExecutor()
		subspace = layer.dataset.subspace(*layer.state.expressions, executor=executor, delay=True)
		means = subspace.mean()
		with dialogs.ProgressExecution(plot_window, "Calculating mean", executor=executor) as progress:
			progress.add_task(means)
			progress.execute()
		logger.debug("get means")
		means = means.get()
		logger.debug("got means")

		vars = subspace.var(means=means)
		with dialogs.ProgressExecution(plot_window, "Calculating variance", executor=executor) as progress:
			progress.add_task(vars)
			progress.execute()
		#limits  = limits.get()
		vars = vars.get()
		stds = vars**0.5
		sigmas = 3
		limits = list(zip(means-sigmas*stds, means+sigmas*stds))
		#plot_window.ranges_show = limits
		plot_window.set_ranges(range(len(limits)), limits, add_to_history=True, reason="3 sigma region")
		#plot_window.update_all_layers()
		#for layer in plot_window.layers:
		#	layer.flag_needs_update()
		logger.debug("means=%r", means)
		logger.debug("vars=%r", vars)
		logger.debug("limits=%r", limits)
		plot_window.queue_history_change("task: 3 sigma region")
		#plot_window.queue_update()

def subtract_mean(plot_window):
	if plot_window.layers:
		layer = plot_window.layers[0]

		executor = vaex.execution.Executor()
		subspace = layer.dataset.subspace(*layer.state.expressions, executor=executor, delay=True)
		means = subspace.mean()
		with dialogs.ProgressExecution(plot_window, "Calculating mean", executor=executor):
			executor.execute()
		means = means.get()
		new_expressions = ["(%s) - %s" % (expression, mean) for expression, mean in zip(layer.state.expressions, means)]
		for i in range(len(new_expressions)):
			vmin, vmax = layer.plot_window.state.ranges_viewport[i]
			vmin -= means[i]
			vmax -= means[i]
			layer.plot_window.set_range(vmin, vmax, i)
		for i in range(len(new_expressions)):
			layer.set_expression(new_expressions[i], i)
		plot_window.update_all_layers()
		plot_window.queue_history_change("task: remove mean")


@vaex.ui.plugin.pluginclass
class TasksPlugin(vaex.ui.plugin.PluginPlot):
	name = "tasks"
	def __init__(self, dialog):
		super(TasksPlugin, self).__init__(dialog)
		dialog.plug_toolbar(self.plug_toolbar, 1.6)


	def plug_toolbar(self):
		logger.info("adding %s plugin" % self.name)

		self.menu = QtGui.QMenu("&Tasks")
		self.dialog.menu_bar.addMenu(self.menu)

		self.tasks_button = QtGui.QToolButton()
		self.tasks_button.setIcon(QtGui.QIcon(iconfile('gear')))
		self.tasks_button.setText("Tasks")
		self.tasks_button.setPopupMode(QtGui.QToolButton.InstantPopup)
		self.tasks_button.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon)
		self.tasks_button.setMenu(self.menu)

		#self.action_tasks = QtGui.QAction(QtGui.QIcon(iconfile('gear')), 'Tasks', self.dialog)
		self.dialog.toolbar.addWidget(self.tasks_button)


		self.action_tasks_loglog = QtGui.QAction(QtGui.QIcon(iconfile('gear')), 'Log-Log', self.dialog)
		self.menu.addAction(self.action_tasks_loglog)
		self.action_tasks_loglog.triggered.connect(lambda *arg: loglog(self.dialog))

		self.action_tasks_loglog = QtGui.QAction(QtGui.QIcon(iconfile('gear')), '3 sigma region', self.dialog)
		self.menu.addAction(self.action_tasks_loglog)
		self.action_tasks_loglog.triggered.connect(lambda *arg: sigma3(self.dialog))

		self.action_tasks_loglog_3sigma = QtGui.QAction(QtGui.QIcon(iconfile('gear')), 'Log-Log and 3 sigma region', self.dialog)
		self.menu.addAction(self.action_tasks_loglog_3sigma)
		self.action_tasks_loglog_3sigma.triggered.connect(lambda *arg: loglog_and_sigma3(self.dialog))

		self.action_tasks_removelog = QtGui.QAction(QtGui.QIcon(iconfile('gear')), 'Remove log', self.dialog)
		self.menu.addAction(self.action_tasks_removelog)
		self.action_tasks_removelog.triggered.connect(lambda *arg: removelog(self.dialog))

		self.action_tasks_subtract_mean = QtGui.QAction(QtGui.QIcon(iconfile('gear')), 'Subtract mean', self.dialog)
		self.menu.addAction(self.action_tasks_subtract_mean)
		self.action_tasks_subtract_mean.triggered.connect(lambda *arg: subtract_mean(self.dialog))


	def __(self):

		self.action_store  = QtGui.QAction(QtGui.QIcon(iconfile('gear')), 'Store', self.dialog)
		self.action_store.setShortcut("Ctrl+B")
		self.action_store_toolbar  = QtGui.QAction(QtGui.QIcon(iconfile('star')), 'Store', self.dialog)


		self.dialog.toolbar.addAction(self.action_store_toolbar)
		self.action_store.triggered.connect(self.on_store)
		self.action_store_toolbar.triggered.connect(self.on_store)
		self.changed_handle = storage_plots.changed.connect(self.load_options_menu)
		self.load_options_menu()

		self.dialog.menu_mode.addSeparator()
		self.action_zoom_rect = QtGui.QAction(QtGui.QIcon(iconfile('zoom')), '&Zoom to rect', self.dialog)
		self.action_zoom_rect.setShortcut("Ctrl+Alt+Z")
		self.dialog.menu_mode.addAction(self.action_zoom_rect)

		self.action_zoom_x = QtGui.QAction(QtGui.QIcon(iconfile('zoom_x')), '&Zoom x', self.dialog)
		self.action_zoom_y = QtGui.QAction(QtGui.QIcon(iconfile('zoom_y')), '&Zoom y', self.dialog)
		self.action_zoom = QtGui.QAction(QtGui.QIcon(iconfile('zoom')), '&Zoom(you should not read this)', self.dialog)

		self.action_zoom_x.setShortcut("Ctrl+Alt+X")
		self.action_zoom_y.setShortcut("Ctrl+Alt+Y")
		self.dialog.menu_mode.addAction(self.action_zoom_x)
		self.dialog.menu_mode.addAction(self.action_zoom_y)


		self.dialog.menu_mode.addSeparator()
		self.action_zoom_out = QtGui.QAction(QtGui.QIcon(iconfile('zoom_out')), '&Zoom out', self.dialog)
		self.action_zoom_in = QtGui.QAction(QtGui.QIcon(iconfile('zoom_in')), '&Zoom in', self.dialog)
		self.action_zoom_fit = QtGui.QAction(QtGui.QIcon(iconfile('arrow_out')), '&Reset view', self.dialog)
		#self.action_zoom_use = QtGui.QAction(QtGui.QIcon(iconfile('chart_bar')), '&Use zoom area', self.dialog)
		self.action_zoom_out.setShortcut("Ctrl+Alt+-")
		self.action_zoom_in.setShortcut("Ctrl+Alt++")
		self.action_zoom_fit.setShortcut("Ctrl+Alt+0")
		self.dialog.menu_mode.addAction(self.action_zoom_out)
		self.dialog.menu_mode.addAction(self.action_zoom_in)
		self.dialog.menu_mode.addAction(self.action_zoom_fit)



		self.dialog.action_group_main.addAction(self.action_zoom_rect)
		self.dialog.action_group_main.addAction(self.action_zoom_x)
		self.dialog.action_group_main.addAction(self.action_zoom_y)

		#self.dialog.toolbar.addAction(self.action_zoom_out)
		#self.dialog.add_shortcut(self.action_zoom_in,"+")
		#self.dialog.add_shortcut(self.action_zoom_out,"-")

		#self.dialog.add_shortcut(self.action_zoom_rect,"Z")
		#self.dialog.add_shortcut(self.action_zoom_x,"Alt+X")
		#self.dialog.add_shortcut(self.action_zoom_y,"Alt+Y")
		#self.dialog.add_shortcut(self.action_zoom_fit, "0")

		self.dialog.toolbar.addAction(self.action_zoom)
		self.zoom_menu = QtGui.QMenu()
		self.action_zoom.setMenu(self.zoom_menu)
		self.zoom_menu.addAction(self.action_zoom_rect)
		self.zoom_menu.addAction(self.action_zoom_x)
		self.zoom_menu.addAction(self.action_zoom_y)
		if self.dialog.dimensions == 1:
			self.lastActionZoom = self.action_zoom_x # this makes more sense for histograms as default
		else:
			self.lastActionZoom = self.action_zoom_rect

		self.dialog.toolbar.addSeparator()
		#self.dialog.toolbar.addAction(self.action_zoom_out)
		self.dialog.toolbar.addAction(self.action_zoom_fit)

		self.action_zoom.triggered.connect(self.onActionZoom)
		self.action_zoom_out.triggered.connect(self.onZoomOut)
		self.action_zoom_in.triggered.connect(self.onZoomIn)
		self.action_zoom_fit.triggered.connect(self.onZoomFit)
		#self.action_zoom_use.triggered.connect(self.onZoomUse)


		self.action_zoom.setCheckable(True)
		self.action_zoom_rect.setCheckable(True)
		self.action_zoom_x.setCheckable(True)
		self.action_zoom_y.setCheckable(True)




from __future__ import print_function
# -*- coding: utf-8 -*-
import math
import collections
import time
import sys
import logging
import psutil
import traceback as tb
import vaex
import smtplib
import platform
import getpass
import os
# import urllib.request, urllib.parse, urllib.error
import vaex.utils
# from email.mime.text import MIMEText

try:
    from urllib.request import urlopen
    from urllib.parse import urlparse, urlencode
except ImportError:
    from urlparse import urlparse
    from urllib import urlopen, urlencode, quote as urlquote

logger = logging.getLogger("vaex.ui.qt")

try:
    import sip
    sip.setapi('QVariant', 2)
    sip.setapi('QString', 2)
    from PyQt4 import QtGui, QtCore, QtTest  # , QtNetwork
    sip.setapi('QVariant', 2)
    # from PyQt4.QtWebKit import QWebView
    qt_version = QtCore.PYQT_VERSION_STR
    qt_mayor = int(qt_version[0])
except ImportError as e1:
    try:
        from PyQt5 import QtGui, QtCore, QtTest, QtWidgets  # , QtNetwork

        for name in dir(QtWidgets):
            if name[0].lower() == "q":
                setattr(QtGui, name, getattr(QtWidgets, name))
        qt_version = QtCore.PYQT_VERSION_STR
        qt_mayor = int(qt_version[0])
        QtGui.QStringListModel = QtCore.QStringListModel
        # QtCore.Slot = QtCore.pyqtSlot
    except ImportError as e1b:
        try:
            from PySide import QtGui, QtCore, QtTest  # , QtNetwork

            # from PySide.QtWebKit import QWebView
            QtCore.pyqtSignal = QtCore.Signal
            qt_version = QtCore.__version__
            qt_mayor = 4
        except ImportError as e2:
            print("could not import PyQt4, PyQt5, or PySide, please install", file=sys.stderr)
            print("errors: ", repr(e1), repr(e2), file=sys.stderr)
            sys.exit(1)


def attrsetter(object, attr_name):
    def setter(value):
        setattr(object, attr_name, value)
    return setter


def attrgetter(object, attr_name):
    def getter():
        return getattr(object, attr_name)
    return getter


class ProgressExecution(object):
    def __init__(self, parent, title, cancel_text="Cancel", executor=None):
        self.parent = parent
        self.title = title
        self.cancel_text = cancel_text
        self.executor = executor
        if self.executor:
            def begin():
                self.time_begin = time.time()
                # self.progress_bar.setValue(0)
                self.cancelled = False
                # self.button_cancel.setEnabled(True)

            def end():
                # self.progress_bar.setValue(1000)
                self.cancelled = False
                # self.button_cancel.setEnabled(False)
                time_total = time.time() - self.time_begin
                # self.label_time.setText("%.2fs" % time_total)

            def progress(fraction):
                # self.progress_bar.setValue(fraction*1000)
                # QtCore.QCoreApplication.instance().processEvents()
                # logger.debug("queue: %r %r", self.queue_update.counter, self.queue_update.counter_processed)
                # return (not self.cancelled) and (not self.queue_update.in_queue(2))
                return self.progress(fraction * 100)

            def cancel():
                # self.progress_bar.setValue(0)
                # self.button_cancel.setEnabled(False)
                # self.label_time.setText("cancelled")
                pass

            # self._begin_signal = self.executor.signal_begin.connect(begin)
            # self._progress_signal = self.executor.signal_progress.connect(progress)
            # self._end_signal = self.executor.signal_end.connect(end)
            # self._cancel_signal = self.executor.signal_cancel.connect(cancel)
        self.tasks = []
        self._task_signals = []

    def execute(self):
        logger.debug("show dialog")
        if isinstance(self.executor, vaex.remote.ServerExecutor):
            self.dialog.exec_()
            logger.debug("dialog stopped")
            # important possible deadlock possible:
            # in the case where fulfill or reject is indirectly called, the call to that promise is put in a qtevent
            # so while they are pending, keep processing events until they are handled in this thread
            while any(task.isPending for task in self.tasks):
                QtCore.QCoreApplication.instance().processEvents()
                time.sleep(0.01)
            self.finished_tasks()
        else:
            self.dialog.show()
            self.executor.execute()
            self.dialog.hide()
        logger.debug("self.dialog.wasCanceled() = %r", self.dialog.wasCanceled())
        return not self.dialog.wasCanceled()

    def add_task(self, task):
        self._task_signals.append(task.signal_progress.connect(self._on_progress))
        self.tasks.append(task)
        return task

    def _on_progress(self, fraction):
        total = self.get_progress_fraction()
        self.progress(total * 100)
        QtCore.QCoreApplication.instance().processEvents()
        ok = not self.dialog.wasCanceled()
        if total == 1:
            self.dialog.hide()
        return ok

    def get_progress_fraction(self):
        total_fraction = 0
        for task in self.tasks:
            total_fraction += task.progress_fraction
        fraction = total_fraction / len(self.tasks)
        return fraction

    def finished_tasks(self):
        for task, signal in zip(self.tasks, self._task_signals):
            task.signal_progress.disconnect(signal)
        self.tasks = []
        self._task_signals = []

    def __enter__(self):
        self.dialog = QtGui.QProgressDialog(self.title, self.cancel_text, 0, 1000, self.parent)
        # self.dialog.show()
        self.dialog.setWindowModality(QtCore.Qt.WindowModal)
        self.dialog.setMinimumDuration(0)
        self.dialog.setAutoClose(True)
        self.dialog.setAutoReset(True)
        # QtCore.QCoreApplication.instance().processEvents()
        return self

    def progress(self, percentage):
        self.dialog.setValue(int(percentage * 10))
        QtCore.QCoreApplication.instance().processEvents()
        # self.dialog.setValue(300)
        # self.dialog.update()
        # self.dialog.repaint()
        # print "progress", `percentage`, type(percentage), int(percentage*10)
        # QtCore.QCoreApplication.instance().processEvents()
        return not self.dialog.wasCanceled()
        # return False
        # raise RuntimeError("progress cancelled")

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.dialog.hide()
        if 0:  # self.executor:
            self.executor.signal_begin.disconnect(self._begin_signal)
            self.executor.signal_progress.disconnect(self._progress_signal)
            self.executor.signal_end.disconnect(self._end_signal)
            self.executor.signal_cancel.disconnect(self._cancel_signal)


class assertError(object):
    def __init__(self, calls_expected=1):
        self.calls_expected = calls_expected

    def wrapper(self, *args, **kwargs):
        self.called += 1

    def __enter__(self):
        global dialog_error
        self.remember = dialog_error
        self.called = 0
        dialog_error = self.wrapper
        logger.debug("wrapped dialog_error")

    def __exit__(self, exc_type, exc_val, exc_tb):
        global dialog_error
        assert self.called == self.calls_expected, "expected the error dialog to be invoked %i time(s), was called %i times(s)" % (self.calls_expected, self.called)
        dialog_error = self.remember
        logger.debug("unwrapped dialog_error")


class settext(object):
    def __init__(self, value, calls_expected=1):
        self.value = value
        self.calls_expected = calls_expected

    def wrapper(self, *args, **kwargs):
        self.called += 1
        return self.value

    def __enter__(self):
        global gettext
        self.remember = gettext
        self.called = 0
        gettext = self.wrapper
        logger.debug("wrapped gettext")

    def __exit__(self, exc_type, exc_val, exc_tb):
        global gettext
        assert self.called == self.calls_expected, "expected the error dialog to be invoked %i time(s), was called %i times(s)" % (self.calls_expected, self.called)
        gettext = self.remember
        logger.debug("unwrapped gettext")


class setchoose(object):
    def __init__(self, value, calls_expected=1):
        self.value = value
        self.calls_expected = calls_expected

    def wrapper(self, *args, **kwargs):
        self.called += 1
        return self.value

    def __enter__(self):
        global choose
        self.remember = choose
        self.called = 0
        choose = self.wrapper
        logger.debug("wrapped choose")

    def __exit__(self, exc_type, exc_val, exc_tb):
        global choose
        assert self.called == self.calls_expected, "expected the error dialog to be invoked %i time(s), was called %i times(s)" % (self.calls_expected, self.called)
        choose = self.remember
        logger.debug("unwrapped choose")


class FakeProgressExecution(object):
    def __init__(self, *args):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def progress(self, percentage):
        return True


class OptionBase(object):
    def setVisible(self, value):
        if value:
            self.show()
        else:
            self.hide()

    def show(self):
        self.combobox.show()
        self.label.show()

    def hide(self):
        self.combobox.hide()
        self.label.hide()


class Codeline(OptionBase):
    def __init__(self, parent, label, options, getter, setter, update=lambda: None):
        self.update = update
        self.options = options
        self.label = QtGui.QLabel(label, parent)
        self.combobox = QtGui.QComboBox(parent)
        self.combobox.addItems(options)
        self.combobox.setEditable(True)

        def wrap_setter(value, update=True):
            self.combobox.lineEdit().setText(value)
            setter(value)
            if update:
                self.update()
        # auto getter and setter
        setattr(self, "get_value", getter)
        setattr(self, "set_value", wrap_setter)

        def on_change(index):
            on_edit_finished()

        def on_edit_finished():
            new_value = text = str(self.combobox.lineEdit().text())
            if new_value != self.current_value:
                self.current_value = new_value
                setter(self.current_value)
                update()
        self.combobox.currentIndexChanged.connect(on_change)
        self.combobox.lineEdit().editingFinished.connect(on_edit_finished)
        # self.combobox.setCurrentIndex(options.index(getter()))
        self.current_value = getter()
        self.combobox.lineEdit().setText(self.current_value)

    def add_to_grid_layout(self, row, grid_layout):
        grid_layout.addWidget(self.label, row, 0)
        grid_layout.addWidget(self.combobox, row, 1)
        return row + 1


class Option(OptionBase):
    def __init__(self, parent, label, options, getter, setter, update=lambda: None):
        self.update = update
        self.options = options
        self.label = QtGui.QLabel(label, parent)
        self.combobox = QtGui.QComboBox(parent)
        self.combobox.addItems(options)

        def wrap_setter(value, update=True):
            self.combobox.setCurrentIndex(options.index(getter()))
            setter(value)
            if update:
                self.update()
        # auto getter and setter
        setattr(self, "get_value", getter)
        setattr(self, "set_value", wrap_setter)

        def on_change(index):
            setter(self.options[index])
            update()
        self.combobox.setCurrentIndex(options.index(getter()))
        self.combobox.currentIndexChanged.connect(on_change)

    def add_to_grid_layout(self, row, grid_layout):
        grid_layout.addWidget(self.label, row, 0)
        grid_layout.addWidget(self.combobox, row, 1)
        return row + 1


class TextOption(OptionBase):
    def __init__(self, parent, label, value, placeholder, getter, setter, update=lambda: None):
        self.update = update
        self.value = value
        self.placeholder = placeholder
        self.label = QtGui.QLabel(label, parent)
        self.textfield = QtGui.QLineEdit(parent)
        self.textfield.setPlaceholderText(self.get_placeholder())
        # self.combobox.addItems(options)

        def wrap_setter(value, update=True):
            self.textfield.setText(value)
            setter(value)
            if update:
                self.update()
        # auto getter and setter
        setattr(self, "get_value", getter)
        setattr(self, "set_value", wrap_setter)

        def on_change(*ignore):
            setter(self.textfield.text())
            update()
        # self.combobox.currentIndexChanged.connect(on_change)
        # self.combobox.setCurrentIndex(options.index(getter()))
        self.textfield.returnPressed.connect(on_change)

        if 1:
            from vaex.ui.icons import iconfile
            self.tool_button = QtGui.QToolButton(parent)
            self.tool_button .setIcon(QtGui.QIcon(iconfile('gear')))
            self.tool_menu = QtGui.QMenu()
            self.tool_button.setMenu(self.tool_menu)
            self.tool_button.setPopupMode(QtGui.QToolButton.InstantPopup)

            if self.placeholder:
                def fill(_=None):
                    value = self.get_placeholder()
                    if value:
                        setter(value)
                        self.set_value(value)
                self.action_fill = QtGui.QAction("Fill in default value", parent)
                self.action_fill.triggered.connect(fill)
                self.tool_menu.addAction(self.action_fill)

            def copy(_=None):
                value = self.textfield.text()
                if value:
                    clipboard = QtGui.QApplication.clipboard()
                    text = str(value)
                    clipboard.setText(text)
            self.action_copy = QtGui.QAction("Copy", parent)
            self.action_copy.triggered.connect(copy)
            self.tool_menu.addAction(self.action_copy)

            def paste(_=None):
                clipboard = QtGui.QApplication.clipboard()
                text = clipboard.text()
                setter(text)
                self.set_value(text)
            self.action_paste = QtGui.QAction("Paste", parent)
            self.action_paste.triggered.connect(paste)
            self.tool_menu.addAction(self.action_paste)

    def set_unit_completer(self):
        self.completer = vaex.ui.completer.UnitCompleter(self.textfield)
        self.textfield.setCompleter(self.completer)

    def get_placeholder(self):
        if callable(self.placeholder):
            return self.placeholder()
        else:
            return self.placeholder

    def add_to_grid_layout(self, row, grid_layout):
        grid_layout.addWidget(self.label, row, 0)
        grid_layout.addWidget(self.textfield, row, 1)
        grid_layout.addWidget(self.tool_button, row, 2)
        return row + 1


class RangeOption(object):
    def __init__(self, parent, label, values, getter, setter, update=lambda: None):
        self.update = update
        self.values = [str(k) for k in values]
        self.label = QtGui.QLabel(label, parent)
        self.combobox_min = QtGui.QComboBox(parent)
        self.combobox_max = QtGui.QComboBox(parent)
        self.combobox_min.setEditable(True)
        self.combobox_max.setEditable(True)
        self.vmin = None
        self.vmax = None
        # self.combobox_min.addItems(values)
        # self.combobox_max.addItems(values)

        def wrap_setter(value, update=True):
            if value is None:
                vmin, vmax = None, None
            else:
                vmin, vmax = value
            self.combobox_min.blockSignals(True)
            self.combobox_max.blockSignals(True)
            changed = False
            if vmin != self.vmin:
                # print(("setting vmin to", vmin))
                self.vmin = vmin
                self.combobox_min.lineEdit().setText(str(self.vmin) if self.vmin is not None else "")
                changed = True
            if vmax != self.vmax:
                # print(( "setting vmax to", vmax))
                self.vmax = vmax
                self.combobox_max.lineEdit().setText(str(self.vmax) if self.vmax is not None else "")
                changed = True
            self.combobox_min.blockSignals(False)
            self.combobox_max.blockSignals(False)
            # self.combobox.setCurrentIndex(options.index(getter()))
            # setter(value)
            if update and changed:
                self.update()
        # auto getter and setter
        setattr(self, "get_value", getter)
        setattr(self, "set_value", wrap_setter)

        def get():
            # vmin, vmax = None, None
            if self.combobox_min.lineEdit().text().strip():
                try:
                    self.vmin = float(self.combobox_min.lineEdit().text())
                except:
                    logger.exception("parsing vmin")
                    dialog_error(self.combobox_min, "Error parsing number", "Cannot parse number: %s" % self.combobox_min.lineEdit().text())

            if self.combobox_max.lineEdit().text().strip():
                try:
                    self.vmax = float(self.combobox_max.lineEdit().text())
                except:
                    logger.exception("parsing vmax")
                    dialog_error(self.combobox_max, "Error parsing number", "Cannot parse number: %s" % self.combobox_max.lineEdit().text())
            return (self.vmin, self.vmax) if self.vmin is not None and self.vmax is not None else None

        def on_change(_ignore=None):
            # setter(self.options[index])
            value = get()
            if value:
                vmin, vmax = value
                if setter((vmin, vmax)):
                    update()
                # self.weight_x_box.lineEdit().editingFinished.connect(lambda _=None: self.onWeightXExpr())

        self.combobox_min.lineEdit().returnPressed.connect(on_change)
        self.combobox_max.lineEdit().returnPressed.connect(on_change)
        # self.combobox_min.currentIndexChanged.connect(on_change)
        # self.combobox_max.currentIndexChanged.connect(on_change)
        self.combobox_layout = QtGui.QHBoxLayout(parent)
        self.combobox_layout.addWidget(self.combobox_min)
        self.combobox_layout.addWidget(self.combobox_max)
        # self.combobox_min.currentIndexChanged.connect(on_change_min)
        # self.combobox_max.currentIndexChanged.connect(on_change_max)
        # self.combobox.setCurrentIndex(options.index(getter()))

        if 1:
            from vaex.ui.icons import iconfile
            self.tool_button = QtGui.QToolButton(parent)
            self.tool_button .setIcon(QtGui.QIcon(iconfile('gear')))
            self.tool_menu = QtGui.QMenu()
            self.tool_button.setMenu(self.tool_menu)
            self.tool_button.setPopupMode(QtGui.QToolButton.InstantPopup)

            def flip(_=None):
                value = get()
                if value:
                    vmin, vmax = value
                    setter((vmax, vmin))
                    self.set_value((vmax, vmin))
            self.action_flip = QtGui.QAction("Flip axis", parent)
            self.action_flip.triggered.connect(flip)
            self.tool_menu.addAction(self.action_flip)

            def copy(_=None):
                value = get()
                if value:
                    clipboard = QtGui.QApplication.clipboard()
                    text = str(value)
                    clipboard.setText(text)
            self.action_copy = QtGui.QAction("Copy", parent)
            self.action_copy.triggered.connect(copy)
            self.tool_menu.addAction(self.action_copy)

            def paste(_=None):
                clipboard = QtGui.QApplication.clipboard()
                text = clipboard.text()
                try:
                    vmin, vmax = eval(text)
                    setter((vmin, vmax))
                    self.set_value((vmin, vmax))
                except Exception as e:
                    dialog_error(parent, "Could not parse min/max values", "Could not parse min/max values: %r" % e)
            self.action_paste = QtGui.QAction("Paste", parent)
            self.action_paste.triggered.connect(paste)
            self.tool_menu.addAction(self.action_paste)

    def add_to_grid_layout(self, row, grid_layout):
        # grid_layout.addLayout(self.combobox_layout, row, 1)
        grid_layout.addWidget(self.label, row, 0)
        grid_layout.addWidget(self.combobox_min, row, 1)
        grid_layout.addWidget(self.tool_button, row, 2)
        row += 1
        # grid_layout.addWidget(self.label, row, 0)
        grid_layout.addWidget(self.combobox_max, row, 1)
        return row + 1


# list of rgb values from tableau20
color_list = [(255, 187, 120),
              (255, 127, 14),
              (174, 199, 232),
              (44, 160, 44),
              (31, 119, 180),
              (255, 152, 150),
              (214, 39, 40),
              (197, 176, 213),
              (152, 223, 138),
              (148, 103, 189),
              (247, 182, 210),
              (227, 119, 194),
              (196, 156, 148),
              (140, 86, 75),
              (127, 127, 127),
              (219, 219, 141),
              (199, 199, 199),
              (188, 189, 34),
              (158, 218, 229),
              (23, 190, 207)]


class ColorOption(object):
    def __init__(self, parent, label, getter, setter, update=lambda: None):
        self.update = update
        # self.options = options
        self.label = QtGui.QLabel(label, parent)
        self.combobox = QtGui.QComboBox(parent)
        index = 0
        self.qt_colors = []
        for color_tuple in color_list:
            # self.combobox.addItems(options)
            self.combobox.addItem(",".join(map(str, color_tuple)))
            model = self.combobox.model().index(index, 0)
            color = QtGui.QColor(*color_tuple)
            self.combobox.model().setData(model, color, QtCore.Qt.BackgroundColorRole)
            index += 1
            self.qt_colors.append(color)

        def wrap_setter(value, update=True):
            index = color_list.index(getter())
            self.combobox.setCurrentIndex(index)
            self.combobox.palette().setColor(QtGui.QPalette.Background, self.qt_colors[index])
            self.combobox.palette().setColor(QtGui.QPalette.Highlight, self.qt_colors[index])
            setter([c / 255. for c in value])
            if update:
                self.update()
        # auto getter and setter
        setattr(self, "get_value", getter)
        setattr(self, "set_value", wrap_setter)

        def on_change(index):
            self.set_value(color_list[index])
            update()
        self.combobox.currentIndexChanged.connect(on_change)
        self.combobox.setCurrentIndex(color_list.index(getter()))

    def add_to_grid_layout(self, row, grid_layout):
        grid_layout.addWidget(self.label, row, 0)
        grid_layout.addWidget(self.combobox, row, 1)
        return row + 1


class Checkbox(object):
    def __init__(self, parent, label_text, getter, setter, update=lambda: None):
        self.update = update
        self.label = QtGui.QLabel(label_text, parent)
        self.checkbox = QtGui.QCheckBox(parent)

        def wrap_setter(value, update=True):
            self.checkbox.setChecked(value)
            setter(value)
            if update:
                self.update()
        # auto getter and setter
        setattr(self, "get_value", getter)
        setattr(self, "set_value", wrap_setter)

        def on_change(state):
            value = state == QtCore.Qt.Checked
            setter(value)
            self.update()
        self.checkbox.setChecked(getter())
        self.checkbox.stateChanged.connect(on_change)

    def add_to_grid_layout(self, row, grid_layout, column_start=0):
        grid_layout.addWidget(self.label, row, column_start + 0)
        grid_layout.addWidget(self.checkbox, row, column_start + 1)
        return row + 1


class Slider(object):
    def __init__(self, parent, label_text, value_min, value_max, value_steps, getter, setter, name=None, format="{0:<0.3f}", transform=lambda x: x, inverse=lambda x: x, update=lambda: None, uselog=False, numeric_type=float):
        if name is None:
            name = label_text
        if uselog:
            def transform(x): return 10**x

            def inverse(x): return math.log10(x)
        # self.properties.append(name)
        self.update = update
        self.label = QtGui.QLabel(label_text, parent)
        self.label_value = QtGui.QLabel(label_text, parent)
        self.slider = QtGui.QSlider(parent)
        self.slider.setOrientation(QtCore.Qt.Horizontal)
        self.slider.setRange(0, value_steps)

        def wrap_setter(value, update=True):
            self.slider.setValue((inverse(value) - inverse(value_min)) / (inverse(value_max) - inverse(value_min)) * value_steps)
            setter(value)
            if update:
                self.update()
        # auto getter and setter
        setattr(self, "get_value", getter)
        setattr(self, "set_value", wrap_setter)

        def update_text():
            # label.setText("mean/sigma: {0:<0.3f}/{1:.3g} opacity: {2:.3g}".format(self.tool.function_means[i], self.tool.function_sigmas[i], self.tool.function_opacities[i]))
            self.label_value.setText(format.format(getter()))

        def on_change(index, slider=self.slider):
            value = numeric_type(index / float(value_steps) * (inverse(value_max) - inverse(value_min)) + inverse(value_min))
            setter(transform(value))
            self.update()
            update_text()
        self.slider.setValue((inverse(getter()) - inverse(value_min)) * value_steps / (inverse(value_max) - inverse(value_min)))
        update_text()
        self.slider.valueChanged.connect(on_change)
        # return label, slider, label_value

    def add_to_grid_layout(self, row, grid_layout):
        grid_layout.addWidget(self.label, row, 0)
        grid_layout.addWidget(self.slider, row, 1)
        grid_layout.addWidget(self.label_value, row, 2)
        return row + 1


class QuickDialog(QtGui.QDialog):
    def __init__(self, parent, title, validate=None):
        QtGui.QDialog.__init__(self, parent)
        self.setWindowTitle(title)
        self.layout = QtGui.QFormLayout(self)
        self.layout.setFieldGrowthPolicy(QtGui.QFormLayout.AllNonFixedFieldsGrow)
        self.values = {}
        self.widgets = {}
        self.validate = validate
        self.button_box = QtGui.QDialogButtonBox(QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel, QtCore.Qt.Horizontal, self)
        self.button_box.accepted.connect(self.check_accept)
        self.button_box.rejected.connect(self.reject)
        self.setLayout(self.layout)

    def get(self):
        self.layout.addWidget(self.button_box)
        if self.exec_() == QtGui.QDialog.Accepted:
            return self.values
        else:
            return None

    def check_accept(self):
        logger.debug("on accepted")
        for name, widget in self.widgets.items():
            if isinstance(widget, QtGui.QLabel):
                pass  # self.values[name] = None
            elif isinstance(widget, QtGui.QLineEdit):
                self.values[name] = widget.text()
            elif isinstance(widget, QtGui.QComboBox):
                self.values[name] = widget.currentText()  # lineEdit().text()
            else:
                raise NotImplementedError
        if self.validate is None or self.validate(self, self.values):
            self.accept()

    def accept(self):  # (self, *args):
        return QtGui.QDialog.accept(self)

    def add_label(self, name, label=""):
        self.widgets[name] = widget = QtGui.QLabel(label)
        self.layout.addRow("", widget)

    def add_text(self, name, label="", value="", placeholder=None):
        self.widgets[name] = widget = QtGui.QLineEdit(value, self)
        if placeholder:
            widget.setPlaceholderText(placeholder)
        self.layout.addRow(label, widget)

    def add_password(self, name, label="", value="", placeholder=None):
        self.widgets[name] = widget = QtGui.QLineEdit(value, self)
        widget.setEchoMode(QtGui.QLineEdit.Password)
        if placeholder:
            widget.setPlaceholderText(placeholder)
        self.layout.addRow(label, widget)

    def add_ucd(self, name, label="", value=""):
        self.widgets[name] = widget = QtGui.QLineEdit(value, self)
        widget.setCompleter(vaex.ui.completer.UCDCompleter(widget))
        self.layout.addRow(label, widget)

    def add_combo_edit(self, name, label="", value="", values=[]):
        self.widgets[name] = widget = QtGui.QComboBox(self)
        widget.addItems([value] + values)
        widget.setEditable(True)
        self.layout.addRow(label, widget)

    def add_expression(self, name, label, value, dataset):
        import vaex.ui.completer
        self.widgets[name] = widget = vaex.ui.completer.ExpressionCombobox(self, dataset)
        if value is not None:
            widget.lineEdit().setText(value)
        self.layout.addRow(label, widget)

    def add_variable_expression(self, name, label, value, dataset):
        import vaex.ui.completer
        self.widgets[name] = widget = vaex.ui.completer.ExpressionCombobox(self, dataset, variables=True)
        if value is not None:
            widget.lineEdit().setText(value)
        self.layout.addRow(label, widget)

    def add_combo(self, name, label="", values=[]):
        self.widgets[name] = widget = QtGui.QComboBox(self)
        widget.addItems(values)
        self.layout.addRow(label, widget)


def get_path_save(parent, title="Save file", path="", file_mask="HDF5 *.hdf5"):
    path = QtGui.QFileDialog.getSaveFileName(parent, title, path, file_mask)
    if isinstance(path, tuple):
        path = str(path[0])  # ]
    return str(path)


def get_path_open(parent, title="Select file", path="", file_mask="HDF5 *.hdf5"):
    path = QtGui.QFileDialog.getOpenFileName(parent, title, path, file_mask)
    if isinstance(path, tuple):
        path = str(path[0])  # ]
    return str(path)


def getdir(parent, title, start_directory=""):
    result = QtGui.QFileDialog.getExistingDirectory(parent, title, "", QtGui.QFileDialog.ShowDirsOnly | QtGui.QFileDialog.DontResolveSymlinks)
    return None if result is None else str(result)


def gettext(parent, title, label, default=""):
    text, ok = QtGui.QInputDialog.getText(parent, title, label, QtGui.QLineEdit.Normal, default)
    return str(text) if ok else None


class Thenner(object):
    def __init__(self, callback):
        self.callback = callback
        self.thennable = False

    def then(self, *args, **kwargs):
        self.thennable = True
        self.args = args
        self.kwargs = kwargs

    def do(self):
        if self.thennable:
            self.callback(*self.args, **self.kwargs)


def set_choose(value, ok=None):
    global choose
    thenner = Thenner(set_choose)

    def wrapper(*args):
        thenner.do()
        return value  # if ok is None else (value, ok)
    choose = wrapper
    return thenner


QtGui.QInputDialog_getItem = QtGui.QInputDialog.getItem


def choose(parent, title, label, options, index=0, editable=False):
    text, ok = QtGui.QInputDialog_getItem(parent, title, label, options, index, editable)
    if editable:
        return text if ok else None
    else:
        return options.index(text) if ok else None


def set_select_many(ok, mask):
    global select_many

    def wrapper(*args):
        return ok, mask
    select_many = wrapper


def select_many(parent, title, options):
    dialog = QtGui.QDialog(parent)
    dialog.setWindowTitle(title)
    dialog.setModal(True)
    layout = QtGui.QGridLayout(dialog)
    dialog.setLayout(layout)

    scroll_area = QtGui.QScrollArea(dialog)
    scroll_area.setWidgetResizable(True)
    frame = QtGui.QWidget(scroll_area)
    layout_frame = QtGui.QVBoxLayout(frame)
    # frame.setMinimumSize(100,400)
    frame.setLayout(layout_frame)
    checkboxes = [QtGui.QCheckBox(option, frame) for option in options]
    scroll_area.setWidget(frame)
    row = 0
    for checkbox in checkboxes:
        checkbox.setCheckState(QtCore.Qt.Checked)
        layout_frame.addWidget(checkbox)  # , row, 0)
        row += 1

    buttonLayout = QtGui.QHBoxLayout()
    button_ok = QtGui.QPushButton("Ok", dialog)
    button_cancel = QtGui.QPushButton("Cancel", dialog)
    button_cancel.clicked.connect(dialog.reject)
    button_ok.clicked.connect(dialog.accept)

    buttonLayout.addWidget(button_cancel)
    buttonLayout.addWidget(button_ok)
    layout.addWidget(scroll_area)
    layout.addLayout(buttonLayout, row, 0)
    # options_selected = []
    value = dialog.exec_()
    mask = [checkbox.checkState() == QtCore.Qt.Checked for checkbox in checkboxes]
    return value == QtGui.QDialog.Accepted, mask


def memory_check_ok(parent, bytes_needed):
    bytes_available = psutil.virtual_memory().available
    alot = bytes_needed / bytes_available > 0.5
    required = vaex.utils.filesize_format(bytes_needed)
    available = vaex.utils.filesize_format(bytes_available)
    msg = "This action required {required} of memory, while you have {available}. Are you sure you want to continue?".format(**locals())
    return not alot or dialog_confirm(parent, "A lot of memory requested", msg)


def dialog_error(parent, title, msg):
    QtGui.QMessageBox.warning(parent, title, msg)


def dialog_info(parent, title, msg):
    QtGui.QMessageBox.information(parent, title, msg)


def dialog_confirm(parent, title, msg, to_all=False):
    # return QtGui.QMessageBox.information(parent, title, msg, QtGui.QMessageBox.Yes|QtGui.QMessageBox.No) == QtGui.QMessageBox.Yes
    msgbox = QtGui.QMessageBox(parent)
    msgbox.setText(msg)
    msgbox.setWindowTitle(title)
    # , title, msg, QtGui.QMessageBox.Yes|QtGui.QMessageBox.No) == QtGui.QMessageBox.Yes
    msgbox.addButton(QtGui.QMessageBox.Yes)
    if to_all:
        msgbox.addButton(QtGui.QMessageBox.YesToAll)
        msgbox.setDefaultButton(QtGui.QMessageBox.YesToAll)
    else:
        msgbox.setDefaultButton(QtGui.QMessageBox.Yes)
    msgbox.addButton(QtGui.QMessageBox.No)
    result = msgbox.exec_()
    if to_all:
        return result in [QtGui.QMessageBox.Yes, QtGui.QMessageBox.YesToAll], result == QtGui.QMessageBox.YesToAll
    else:
        return result in [QtGui.QMessageBox.Yes]


confirm = dialog_confirm


def email(text):
    osname = platform.system().lower()
    if osname == "linux":
        text = text.replace("#", "%23")  # for some reason, # needs to be double quoted on linux, otherwise it is interpreted as comment symbol

    body = urlquote(text)

    subject = urlquote('Error report for: ' + vaex.__full_name__)
    mailto = "mailto:maartenbreddels@gmail.com?subject={subject}&body={body}".format(**locals())
    print("open:", mailto)
    vaex.utils.os_open(mailto)


def old_email(text):
    # Open a plain text file for reading.  For this example, assume that
    msg = MIMEText(text)

    msg['Subject'] = 'Error report for: ' + vaex.__full_name__
    email_from = "vaex@astro.rug.nl"
    # email_to = "breddels@astro.rug.nl"
    email_to = "maartenbreddels@gmail.com"
    msg['From'] = email_to
    msg['To'] = email_to

    # Send the message via our own SMTP server, but don't include the
    # envelope header.
    # s = smtplib.SMTP('mailhost.astro.rug.nl')
    s = smtplib.SMTP('smtp.gmail.com')
    s.helo('fw1.astro.rug.nl')
    s.sendmail(email_to, [email_to], msg.as_string())
    s.quit()


def qt_exception(parent, exctype, value, traceback):
    trace_lines = tb.format_exception(exctype, value, traceback)
    trace = "".join(trace_lines)
    print(trace)
    info = "username: %r\n" % (getpass.getuser(),)
    info += "program: %r\n" % vaex.__program_name__
    info += "version: %r\n" % vaex.__version__
    info += "full name: %r\n" % vaex.__full_name__
    info += "arguments: %r\n" % sys.argv
    info += "Qt version: %r\n" % qt_version

    attrs = sorted(dir(platform))
    for attr in attrs:
        if not attr.startswith("_") and attr not in ["popen", "system_alias"]:
            f = getattr(platform, attr)
            if isinstance(f, collections.Callable):
                try:
                    info += "%s: %r\n" % (attr, f())
                except:
                    pass
    # , platform.architecture(), platform.dist(), platform.linux_distribution(),

    report = info + "\n" + trace
    text = """An unexpected error occured, you may press ok and continue, but the program might be unstable.

""" + report

    dialog = QtGui.QMessageBox(parent)
    dialog.setText("Unexpected error: %s\nDo you want to continue" % (exctype, ))
    # dialog.setInformativeText(text)
    dialog.setDetailedText(text)
    buttonSend = QtGui.QPushButton("Email report", dialog)
    buttonQuit = QtGui.QPushButton("Quit program", dialog)
    buttonContinue = QtGui.QPushButton("Continue", dialog)

    def exit(ignore=None):
        print("exit")
        sys.exit(1)

    def _email(ignore=None):
        if QtGui.QMessageBox.information(dialog, "Send report", "Confirm that you want to send a report", QtGui.QMessageBox.Abort | QtGui.QMessageBox.Yes) == QtGui.QMessageBox.Yes:
            email(report)
    buttonQuit.clicked.connect(exit)
    buttonSend.clicked.connect(_email)
    dialog.addButton(buttonSend, QtGui.QMessageBox.YesRole)
    dialog.addButton(buttonQuit, QtGui.QMessageBox.NoRole)
    dialog.addButton(buttonContinue, QtGui.QMessageBox.YesRole)
    dialog.setDefaultButton(buttonSend)
    dialog.setEscapeButton(buttonContinue)
    dialog.raise_()
    dialog.exec_()
    # QtGui.QMessageBox.critical(parent, "Unexpected error: %r" % (value, ), text)



__author__ = 'maartenbreddels'

import numpy as np

import vaex.ui.plugin
from vaex.ui.qt import *
import vaex.ui.plot_windows
import logging


logger = logging.getLogger("plugin.zoom")


@vaex.ui.plugin.pluginclass
class Vector3dPlugin(vaex.ui.plugin.PluginLayer):
	name = "vector3"
	def __init__(self, parent, layer):
		super(Vector3dPlugin, self).__init__(parent, layer)
		layer.plug_page(self.plug_page, "Vector field", 2., 2.)

	@staticmethod
	def useon(dialog_class):
		return issubclass(dialog_class, vaex.ui.plot_windows.VolumeRenderingPlotDialog)

	def plug_page(self, page):
		existing_layout = page.layout()
		if isinstance(existing_layout, QtGui.QGridLayout):
			layout = existing_layout
		else:
			#raise NotImplementedError("expected different layout")
			self.layout = layout = QtGui.QGridLayout()
			existing_layout.addLayout(self.layout)
		page.setLayout(self.layout)
		layout.setSpacing(0)
		layout.setContentsMargins(0,0,0,0)
		layout.setAlignment(QtCore.Qt.AlignTop)

		row = 0


		if 0:
			def setter(value):
				self.layer.plot_window.widget_volume.draw_vectors = value
				self.layer.plot_window.widget_volume.update()
			self.vector3d_show_checkbox = self.layer.plot_window.create_checkbox(page, "show 3d vectors", lambda : self.layer.plot_window.widget_volume.draw_vectors, setter)
			layout.addWidget(self.vector3d_show_checkbox, row, 1)
			row += 1


		self.vector3d_show_checkbox = Checkbox(page, "show 3d vectors", getter=attrgetter(self.layer.plot_window.widget_volume, "draw_vectors"), setter=attrsetter(self.layer.plot_window.widget_volume, "draw_vectors"), update=self.layer.plot_window.widget_volume.update)
		row = self.vector3d_show_checkbox.add_to_grid_layout(row, layout)



		self.vector3d_auto_scale_checkbox = Checkbox(page, "auto scale 3d vectors", getter=attrgetter(self.layer.plot_window.widget_volume, "vector3d_auto_scale"), setter=attrsetter(self.layer.plot_window.widget_volume, "vector3d_auto_scale"), update=self.layer.plot_window.widget_volume.update)
		row = self.vector3d_auto_scale_checkbox.add_to_grid_layout(row, layout)


		self.vector3d_min_level_label = Slider(page, "min level 3d", 0, 1, 1000, getter=attrgetter(self.layer.plot_window.widget_volume, "min_level_vector3d"),
		                                     setter=attrsetter(self.layer.plot_window.widget_volume, "min_level_vector3d"), update=self.layer.plot_window.widget_volume.update)
		row = self.vector3d_min_level_label.add_to_grid_layout(row, layout)

		self.vector3d_max_level_label = Slider(page, "max level 3d", 0, 1, 1000, getter=attrgetter(self.layer.plot_window.widget_volume, "max_level_vector3d"),
		                                     setter=attrsetter(self.layer.plot_window.widget_volume, "max_level_vector3d"), update=self.layer.plot_window.widget_volume.update)
		row = self.vector3d_max_level_label.add_to_grid_layout(row, layout)
		return

		def setter(value):
			self.layer.plot_window.widget_volume.vector3d_auto_scale = value
			self.layer.plot_window.widget_volume.update()
		self.vector3d_auto_scale_checkbox = self.layer.plot_window.create_checkbox(page, "auto scale 3d vectors", lambda : self.layer.plot_window.widget_volume.vector3d_auto_scale, setter)
		layout.addWidget(self.vector3d_auto_scale_checkbox, row, 1)
		row += 1


		def setter(value):
			self.layer.plot_window.widget_volume.min_level_vector3d = value
			self.layer.plot_window.widget_volume.update()
		self.vector3d_min_level_label, self.vector3d_min_level_slider, self.vector3d_min_level_value_label =\
				self.layer.plot_window.create_slider(page, "min level: ", 0., 1., lambda : self.layer.plot_window.widget_volume.min_level_vector3d, setter)
		layout.addWidget(self.vector3d_min_level_label, row, 0)
		layout.addWidget(self.vector3d_min_level_slider, row, 1)
		layout.addWidget(self.vector3d_min_level_value_label, row, 2)
		row += 1

		def setter(value):
			self.layer.plot_window.widget_volume.max_level_vector3d = value
			self.layer.plot_window.widget_volume.update()
		self.vector3d_max_level_label, self.vector3d_max_level_slider, self.vector3d_max_level_value_label =\
				self.layer.plot_window.create_slider(page, "max level: ", 0., 1., lambda : self.layer.plot_window.widget_volume.max_level_vector3d, setter)
		layout.addWidget(self.vector3d_max_level_label, row, 0)
		layout.addWidget(self.vector3d_max_level_slider, row, 1)
		layout.addWidget(self.vector3d_max_level_value_label, row, 2)
		row += 1

		def setter(value):
			self.layer.plot_window.widget_volume.vector3d_scale = value
			self.layer.plot_window.widget_volume.update()
		self.vector3d_scale_level_label, self.vector3d_scale_level_slider, self.vector3d_scale_level_value_label =\
				self.layer.plot_window.create_slider(page, "scale: ", 1./20, 20., lambda : self.layer.plot_window.widget_volume.vector3d_scale, setter, format=" {0:>05.2f}", transform=lambda x: 10**x, inverse=lambda x: np.log10(x))
		layout.addWidget(self.vector3d_scale_level_label, row, 0)
		layout.addWidget(self.vector3d_scale_level_slider, row, 1)
		layout.addWidget(self.vector3d_scale_level_value_label, row, 2)
		row += 1

		layout.setRowMinimumHeight(row, 8)
		row += 1




import pylab
import numpy as np
import os
import json
import sys

name = "{name}" if len(sys.argv) == 1 else sys.argv[1]

filename_grid = name + "_grid.npy"
filename_grid_vector = name + "_grid_vector.npy"
filename_grid_selection = name + "_grid_selection.npy"
filename_meta = name + "_meta.json"

has_selection = os.path.exists(filename_grid_selection)
has_vectors = os.path.exists(filename_grid_vector)

meta = json.load(file(filename_meta))
grid = np.load(filename_grid)

xmin, xmax, ymin, ymax = meta["extent"]

if has_selection:
    grid_selection = np.load(filename_grid_selection)
    pylab.imshow(grid, extent=meta["extent"], origin="lower", cmap="binary")
    pylab.imshow(grid_selection, alpha=0.4, extent=meta["extent"], origin="lower")
else:
    pylab.imshow(grid, extent=meta["extent"], origin="lower")

if has_vectors:
    vx, vy, vz = np.load(filename_grid_vector)
    N = vx.shape[0]
    dx = (xmax - xmin) / float(N)
    dy = (ymax - ymin) / float(N)
    x = np.linspace(xmin + dx / 2, xmax - dx / 2, N)
    y = np.linspace(ymin + dy / 2, ymax - dy / 2, N)
    x2d, y2d = np.meshgrid(x, y)
    pylab.quiver(x2d, y2d, vx, vy, color="white")

pylab.show()



__author__ = 'maartenbreddels'

import functools

import matplotlib.patches as patches
import numpy as np
import matplotlib.artist as artist

import vaex.ui.plugin
from vaex.ui.qt import *
import logging


logger = logging.getLogger("plugin.dispersions")


import matplotlib.transforms as transforms
from matplotlib.path import Path

class DispersionEllipse(patches.Patch):
	"""
	This ellipse has it's center in user coordinates, and the width and height in device coordinates
	such that is is not deformed
	"""
	def __str__(self):
		return "DispersionEllipse(%s,%s;%sx%s)" % (self.center[0], self.center[1],
										 self.width, self.height)

	#@docstring.dedent_interpd
	def __init__(self, xy, width, height, scale=1.0, angle=0.0, **kwargs):
		"""
		*xy*
		  center of ellipse

		*width*
		  total length (diameter) of horizontal axis

		*height*
		  total length (diameter) of vertical axis

		*angle*
		  rotation in degrees (anti-clockwise)

		Valid kwargs are:
		%(Patch)s
		"""
		patches.Patch.__init__(self, **kwargs)

		self.center = xy
		self.width, self.height = width, height
		self.scale = scale
		self.angle = angle
		self._path = Path.unit_circle()
		# Note: This cannot be calculated until this is added to an Axes
		self._patch_transform = transforms.IdentityTransform()

	def _recompute_transform(self):
		"""NOTE: This cannot be called until after this has been added
				 to an Axes, otherwise unit conversion will fail. This
				 maxes it very important to call the accessor method and
				 not directly access the transformation member variable.
		"""
		center = (self.convert_xunits(self.center[0]),
				  self.convert_yunits(self.center[1]))
		width = self.width #self.convert_xunits(self.width)
		height = self.height #self.convert_yunits(self.height)
		trans = artist.Artist.get_transform(self)
		self._patch_transform = transforms.Affine2D() \
			.scale(width * 0.5 * self.scale, height * 0.5* self.scale) \
			.rotate_deg(self.angle) \
			.translate(*trans.transform(center))

	def get_path(self):
		"""
		Return the vertices of the rectangle
		"""
		return self._path

	def get_transform(self):
		"""
		Return the :class:`~matplotlib.transforms.Transform` applied
		to the :class:`Patch`.
		"""
		return self.get_patch_transform()

	def get_patch_transform(self):
		self._recompute_transform()
		return self._patch_transform

	def contains(self, ev):
		if ev.x is None or ev.y is None:
			return False, {}
		x, y = self.get_transform().inverted().transform_point((ev.x, ev.y))
		return (x * x + y * y) <= 1.0, {}


class DispersionPlugin(vaex.ui.plugin.PluginLayer):
	name = "dispersion"
	def __init__(self, parent, layer):
		super(DispersionPlugin, self).__init__(parent, layer)
		dialog.plug_page(self.plug_page, "Dispersions", 2.25, 1.0)
		dialog.plug_grids(self.define_grids, self.draw_grids)

	def define_grids(self, grids):
		#grids.define_grid("counts_vector", self.dialog.gridsize_vector, "VZ*0+1")

		# covariance matrix terms
		# diagonals
		for dimension in range(self.dialog.dimensions):
			axis_name = self.dialog.axisnames[dimension].lower()
			expression = self.expressions[dimension].strip()
			if len(expression) > 0:
				grids.define_grid(axis_name + "_mom1", self.dialog.vector_grid_size, expression)
				grids.define_grid(axis_name + "_mom2", self.dialog.vector_grid_size, "(" + expression + ")**2")
			else:
				grids.define_grid(axis_name + "_mom1", self.dialog.vector_grid_size, None)
				grids.define_grid(axis_name + "_mom2", self.dialog.vector_grid_size, None)

		if 1:
			for dimension1 in range(self.dialog.dimensions):
				for dimension2 in range(dimension1+1, self.dialog.dimensions):
					axis_name1 = self.dialog.axisnames[dimension1].lower()
					axis_name2 = self.dialog.axisnames[dimension2].lower()
					expression1 = self.expressions[dimension1].strip()
					expression2 = self.expressions[dimension2].strip()
					if len(expression1) > 0 and  len(expression2) > 0:
						grids.define_grid("cov_" + axis_name1 +"_" +axis_name2, self.dialog.vector_grid_size, "(" + expression1 + ")*(" + expression2 +")")

	def draw_grids(self, axes, grid_map, grid_map_vector):
		if not self.dispersions_draw:
			return
		self.ellipses = []
		dispersions = []
		counts = grid_map_vector["counts"]
		#print "counts check", np.sum(counts), np.sum(grid_map["counts"])
		#print counts
		#print grid_map_vector.keys()
		if self.dialog.dimensions == 2:
			axis_name1 = self.dialog.axisnames[0].lower()
			axis_name2 = self.dialog.axisnames[1].lower()
			if len(self.expressions[0]) > 0 and len(self.expressions[1]) > 0:
				meanx = grid_map_vector[axis_name1 + "_mom1"]/counts
				meany = grid_map_vector[axis_name2 + "_mom1"]/counts
				varx  = grid_map_vector[axis_name1 + "_mom2"]/counts
				vary  = grid_map_vector[axis_name2 + "_mom2"]/counts
				covxy = grid_map_vector["cov_" +axis_name1 + "_" +axis_name2]/counts - meanx*meany
				sigmax = (varx-meanx**2)**0.5
				sigmay = (vary-meany**2)**0.5
				mask = counts > 0
				x = grid_map_vector["x"]
				y = grid_map_vector["y"]
				x, y = np.meshgrid(x, y)


				vmax = np.nanmax(np.sqrt(sigmax.reshape(-1)**2 + sigmay.reshape(-1)**2))

				width, height = self.dialog.canvas.get_width_height()
				#print "width,height", width, height
				max_size = min(width, height) / float(self.dialog.vector_grid_size)# * 0.9
				#print max_size
				#identity_transform = matplotlib.transforms.IdentityTransform()
				#deltax = self.dialog.ranges_show[0][1] - self.dialog.ranges_show[0][0]
				#deltay = self.dialog.ranges_show[1][1] - self.dialog.ranges_show[1][0]
				#aspect = deltay / float(height) / (deltax/float(width))
				#for grid in [x, y, sigmax, sigmay, covxy, counts, mask]:
				#	print grid.shape
				for x, y, sigmax, sigmay, covxy in zip(x[mask].reshape(-1), y[mask].reshape(-1), sigmax[mask].reshape(-1), sigmay[mask].reshape(-1), covxy[mask].reshape(-1)):
					try:
						covmatrix = [[sigmax**2, covxy], [covxy, sigmay**2]]
						eigen_values, eigen_vectors = np.linalg.eig(covmatrix)
					except:
						pass
					else:
						scaling = 1./vmax * max_size
						device_width =  (np.sqrt(np.max(eigen_values)) * scaling)
						device_height = (np.sqrt(np.min(eigen_values)) * scaling)
						if self.dispersions_unit_length:
							length = np.sqrt(device_width**2+device_height**2)
							device_width  /= float(length) / max_size
							device_height /= float(length) / max_size
						#ellipse_width = np.sqrt(np.max(eigen_values)) * scaling / width * deltax
						#ellipse_height = np.sqrt(np.min(eigen_values)) * scaling / height * deltay
						#ellipse_height /= aspect
						if sigmax < sigmay: # if x was smaller, the largest eigenvalue corresponds to the y value
							device_width, device_height = device_height, device_width
							#ellipse_width, ellipse_height = ellipse_height, ellipse_width
						#ellipse_height /= aspect
						angle = np.arctan(2*covxy / (sigmax**2-sigmay**2))/2.
						#angle2 = np.arctan(2*covxy / (sigmax**2-sigmay**2))/2.
						#angle = angle2 = 0
						#print aspect, sigmax, sigmay, sigmax/sigmay, covxy/(sigmax*sigmay), ellipse_width/ellipse_height
						#aspect = 0.1
						#m = [[np.cos(angle2), np.sin(angle2)*aspect], [-np.sin(angle2), np.cos(angle2)*aspect]]
						#ellipse_width, ellipse_height = np.dot(m, [ellipse_width, ellipse_height])
						#print covxy/(sigmax*sigmay), angle, sigmax, sigmay, covxy
						#device_x, device_y = axes.transData.transform((x, y))
						#print device_x, device_y, device_width, device_height
						#ellipse = patches.Ellipse(xy=(device_x, device_y), width=device_width, height=device_height, angle=angle, transform=identity_transform,
						#                         alpha=0.4, color="blue") #rand()*360
						#ellipse = patches.Ellipse(xy=(x, y), width=ellipse_width, height=ellipse_height, angle=np.degrees(angle),
						#                         alpha=0.4, color="blue") #rand()*360
						ellipse = DispersionEllipse(xy=(x, y), width=device_width, height=device_height, angle=np.degrees(angle), scale=self.scale_dispersion,
												  alpha=0.4, facecolor="green", edgecolor="black") #rand()*360

						axes.add_artist(ellipse)
						self.ellipses.append(ellipse)
						#axes.quiver()

						#[Ellipse(xy=rand(2)*10, width=rand(), height=rand(), angle=rand()*360)




	#@staticmethod
	#def useon(dialog_class):
	#	return issubclass(dialog_class, vaex.plot_windows.VolumeRenderingPlotDialog)

	def plug_page(self, page):
		layout = self.layout = QtGui.QGridLayout()
		page.setLayout(self.layout)
		layout.setSpacing(0)
		layout.setContentsMargins(0,0,0,0)
		layout.setAlignment(QtCore.Qt.AlignTop)


		row = 0

		self.dispersions_draw = bool(eval(self.dialog.options.get("disp_draw", "True")))
		def setter(value):
			self.dispersions_draw = value
			self.dialog.plot()
		self.dispersions_draw_checkbox = self.dialog.create_checkbox(page, "Draw dispersion tensors", lambda : self.dispersions_draw, setter)
		layout.addWidget(self.dispersions_draw_checkbox, row, 1)
		row += 1

		self.dispersions_unit_length = bool(eval(self.dialog.options.get("disp_unit", "False")))
		def setter(value):
			self.dispersions_unit_length = value
			self.dialog.plot()
		self.dispersions_unit_lengthcheckbox = self.dialog.create_checkbox(page, "Unit length", lambda : self.dispersions_unit_length, setter)
		layout.addWidget(self.dispersions_unit_lengthcheckbox, row, 1)
		row += 1


		self.expressions = []
		self.expression_boxes = []


		for dimension in range(self.dialog.dimensions):
			axis_name = self.dialog.axisnames[dimension]
			expression_box = QtGui.QComboBox(page)
			expression_box.setEditable(True)
			expression_box.setMinimumContentsLength(10)
			self.expression_boxes.append(expression_box)

			self.layout.addWidget(QtGui.QLabel(axis_name + '-axis:', page), row, 0)
			self.layout.addWidget(expression_box, row, 1, QtCore.Qt.AlignLeft)

			expression = self.dialog.options.get("disp"+axis_name.lower(), "")
			expression_box.lineEdit().setText(expression)
			self.expressions.append(expression)
			#self.onExpressionChangedPartials.append()
			#expression_box.lineEdit().editingFinished.connect(self.onExpressionChangedPartials[axisIndex])
			calllback = functools.partial(self.onExpressionChanged, axis_index=dimension)
			expression_box.lineEdit().editingFinished.connect(calllback)
			row += 1

		self.scale_dispersion = eval(self.dialog.options.get("disp_scale", "1"))
		def setter(value):
			self.scale_dispersion = value
			for ellipse in self.ellipses:
				ellipse.scale = self.scale_dispersion
			self.dialog.canvas.draw()
			#self.dialog.plot()
		self.scale_dispersion_label, self.scale_dispersion_slider, self.scale_dispersion_value_label =\
				self.dialog.create_slider(page, "scale: ", 1./100, 100., lambda : self.scale_dispersion, setter, format=" {0:>05.2f}", transform=lambda x: 10**x, inverse=lambda x: np.log10(x))
		layout.addWidget(self.scale_dispersion_label, row, 0)
		layout.addWidget(self.scale_dispersion_slider, row, 1)
		layout.addWidget(self.scale_dispersion_value_label, row, 2)
		row += 1


	def onExpressionChanged(self, _=None, axis_index=-1):
		text = str(self.expression_boxes[axis_index].lineEdit().text())
		logger.debug("text set for axis %i: %s" % (axis_index, text))
		if text != self.expressions[axis_index]:
			axis_name = self.dialog.axisnames[axis_index].lower()
			self.expressions[axis_index] = text
			if text == "": # check if we can replot without doing the whole calculation
				self.dialog.plot()
			else:
				non_empty = [k for k in self.expressions if len(k) > 0]
				if len(non_empty) == len(self.expressions):
					self.dialog.compute()
					self.dialog.jobsManager.execute()
		else:
			logger.debug("nothing changed")


